#include "common.h"
#include "functions.h"


uint32 hasDot;					// slot or method if set
uint32 dotIndex;				// position of slot or method dot
						// or colon position in cast
uint32 hasColon;				// cast if set
struct type_node *type_head, *type_tail;	// type list


//////////////////////////////////////////////////////////////////////////


FILE *find_api( uint32 uuid ){
	struct type_node *node;
	char filename[500];

	node = type_head;
	while(node != NULL){
		if(node->uuid == uuid){
			sprintf(filename, "%s/%s/api", TYPE_PATH, node->name);
			if((node->api = fopen(filename, "r")) == NULL){
				printf("ERROR: unable to open API file, %s", filename);
				return(0);
			}
			return(node->api);
		}
		node = node->next;
	}
	return(NULL);
}


//////////////////////////////////////////////////////////////////////////


uint32 load_db_types( ){
	struct type_node *node;
	char filename[500], typename[256];
	FILE *typedb;
	uint32 uuid;

	type_head = NULL;

	strcpy(filename, TYPE_PATH);
	strcat(filename, "/type.db");
	if((typedb = fopen(filename, "r")) == NULL) return(0);

	while(!feof(typedb)){
		if(fscanf(typedb, "%d %s\n", &uuid, typename) != 2) break;
		if((node = malloc(sizeof(struct type_node)+strlen(typename)+1)) == NULL){
			printf("ERROR: malloc failed in load_db_types\n");
			fclose(typedb);
			return(0);
		}
		node->next = NULL;
		node->api = NULL;
		node->uuid = uuid;
		node->name_size = strlen(typename);
		strcpy(node->name, typename);
		if(type_head == NULL){
			type_head = node;
			type_tail = node;
		} else {
			type_tail->next = node;
			type_tail = node;
		}
	}
	fclose(typedb);
	return(1);
}


uint32 stringIsType( char *name, uint32 *uuid ){
	struct type_node *node;
	uint32 size;

	size = strlen(name);
	node = type_head;
	while(node != NULL){
		if(node->name_size == size) if(stringCompareData(node->name, name)){
			*uuid = node->uuid;
			return(1);
		}
		node = node->next;
	}
	return(0);
}


uint32 stringIsVar( char *name, uint32 *var, uint32 *uuid ){
	struct var_node *node;
	int32 size, a;

	size = strlen(name);
	hasDot = 0;
	hasColon = 0;
	for(a=0; a<size; a++){
		if(name[a] == '.'){
			hasDot = 1;
			dotIndex = a + 1;
			name[a] = '\0';
			size = strlen(name);
			break;
		} else if(name[a] == ':'){
			hasColon = 1;
			dotIndex = a + 1;
			name[a] = '\0';
			size = strlen(name);
			break;
		}
	}
	a = 0;
	node = var_head;
	while(node != NULL){
		if(node->name_size == size) if(stringCompareData(node->name, name)){
			*var = a;
			*uuid = node->uuid;
			return(1);
		}
		node = node->next;
		a++;
	}
	return(0);
}


uint32 stringIsSlot( char *name, uint32 uuid, uint32 *slot ){
	FILE *api;
	char slotname[256], typename[256];

	if((api = find_api(uuid)) == NULL) return(0);
	fseek(api, 0, SEEK_SET);
	while(!feof(api)){
		if(fscanf(api, "%d %s %s\n", slot, slotname, typename) != 3) break;
		if(*slot == -1) return(0);
		if(stringCompareData(slotname, name)) return(1);
	}
	return(0);
}


uint32 stringIsMethod( char *name, uint32 uuid, uint32 *method, uint32 *return_uuid ){
	FILE *api;
	char methodname[256], typename[256];

	if((api = find_api(uuid)) == NULL) return(0);
	fseek(api, 0, SEEK_SET);
	while(!feof(api)){
		// skip the slot descriptions
		if(fscanf(api, "%d %s %s\n", method, methodname, typename) != 3) break;
		if(*method == -1) break;
	}
	while(!feof(api)){
		// find the method ID
		if(fscanf(api, "%d %s %s\n", method, methodname, typename) != 3) break;
		if(stringCompareData(methodname, name)){

// lookup return type for method

			*return_uuid = 0xffffffff;
			return(1);
		}
	}
	return(0);
}
