#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


void castDisplayModifiers( char *out, s_modifiers mods ){
	if(mods.constant1) strcat(out, "!");
	if(mods.pointer) strcat(out, "@");    
	if(mods.lambda) strcat(out, "&");
	if(mods.array) strcat(out, "#");
	if(mods.unpacked) strcat(out, "[");
	if(mods.constant3) strcat(out, "!");  
	if(mods.dpointer) strcat(out, "@");
	if(mods.dconstant) strcat(out, "!");
	if(mods.vartype) strcat(out, "=");
}


////////////////////////////////////////////////////////////////////////// GLOBAL


s_cast *castCreate( s_string *s ){
	uint32 i, c, level;
	s_cast *item;

	i = 0;
	if(stringFindChar(s, &i, ':')){
		item = (s_cast *)malloc(sizeof(s_cast));
		if(item == NULL){
			printf("ERROR: malloc error in castCreate()\n");
			return(NULL);
		}

		// get name
		item->name = stringCloneRange(s, 0, i - 1);
		i++;

		// collect modifiers
		item->mods.all = 0;
		level = 0;
		while(i != s->size && level < 5){
			switch(s->data[i]){
			case '!': // constant
				item->mods.x.dconstant = 1;
				i++;
				break;
			case '@': // pointer
				if(item->mods.x.dconstant){
					item->mods.x.dconstant = 0;
					c = 1;
				} else c = 0;

				if(level == 0){
					if(c) item->mods.x.constant1 = 1;
					item->mods.x.pointer = 1;
					level = 1;
				} else if(level == 2){
					if(c) item->mods.x.constant3 = 1;
					item->mods.x.dpointer = 1;
					level = 3;
				} else return(NULL);
				i++;
				break;
			case '&': // lambda
				if(level >= 2) return(NULL);
				item->mods.x.lambda = 1;
				level = 2;
				i++;
				break;
			case '#': // array
				if(level >= 2) return(NULL);
				item->mods.x.array = 1;
				level = 2;
				i++;
				break;
			case '[': // unpacked
				if(level >= 2) return(NULL);
				item->mods.x.unpacked = 1;
				level = 2;
				i++;
				break;
			case '=': // variable type
				item->mods.x.vartype = 1; 
				level = 5;
				i++;
				break;
			default: level = 6;
			}
		}
		// get type
		item->type = stringCloneRange(s, i, s->size - 1);
		return(item);
	}

	return(NULL);
}


void castFree( s_cast *cast ){
	stringFree(cast->name);
	stringFree(cast->type);
	free(cast);
}


uint32 castDisplay( char *out, s_node *node ){
	s_cast *cast;

	if(node->type == T_CAST){
		cast = (s_cast *)node->node;
		strcat(out, cast->name->data);			// name:???type
		strcat(out, ":");
		castDisplayModifiers(out, cast->mods.x);
		strcat(out, cast->type->data);
		if(node->next != NULL)
			strcat(out, " ");				// space
	} else if(node->type == T_LIST_CODE && ((s_node*)node->node)->type == T_CAST){
		strcat(out, "(");				// opening paren
		node = node->node;
		cast = (s_cast *)node->node;
		strcat(out, cast->name->data);			// name:????type
		strcat(out, ":");
		castDisplayModifiers(out, cast->mods.x);
		strcat(out, cast->type->data);
		strcat(out, " ");
// output each parameter
strcat(out, "ERR");
		strcat(out, ")");				// closing paren and space
		if(node->next != NULL)
			strcat(out, " ");			// space
	} else return(0);
	return(1);
}
