#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdarg.h>


#define __DEBUG__


#define FILE_DICTIONARY "dictionary.asm"
#define FILE_TEMP ".function.tmp"


//
// main.c
//

#ifndef uint32
	typedef unsigned int uint32;
#endif
#ifndef int32
	typedef int int32;
#endif
#ifndef uint64
	typedef unsigned long long uint64;
#endif
#ifndef int64
	typedef long long int64;
#endif

extern struct s_type *void_type;


//
// parse.c
//


//
// type.c
//


//
// node.c
//

typedef enum e_node_type {
	T_LIST_OUTTER,
	T_LIST_CODE, T_LIST_LAMBDA, T_LIST_DATA,
	T_COMMENT, T_STRING, T_SYMBOL, T_USTRING,
	T_NUMBER,
	T_TYPE, T_CAST
} e_node_type;

typedef struct s_node {
	struct s_node *next;
	enum e_node_type type;
	struct s_type *eval_type;
	void *node;
} s_node;


//
// string.c
//

typedef struct s_string {
	uint32 size;
	char *data;
} s_string;


//
// number.c
//

typedef enum e_num_type {
	NUMBER_FLOAT,
	NUMBER_SIGNED,
	NUMBER_OK,
	NOT_A_NUMBER,
	NUMBER_INVALID,
	NUMBER_TOO_BIG
} e_num_type;

typedef struct s_num {
	enum e_num_type type;
	uint64 value;
} s_num;


//
// cast.c
//

typedef struct s_modifiers {
	uint32 constant1:1;		// 1
	uint32 pointer:1;		// 1
	uint32 lambda:1;		// 2
	uint32 array:1;			// 2
	uint32 unpacked:1;		// 2
	uint32 constant3:1;		// 3
	uint32 dpointer:1;		// 3
	uint32 dconstant:1;		// 4
	uint32 vartype:1;		// 5
} s_modifiers;

typedef struct s_cast {
	struct s_string *name;
	union {
		s_modifiers x;
		uint32 all;
	} mods;
	struct s_string *type;
} s_cast;
typedef struct s_type {
	union {
		s_modifiers x;
		uint32 all;
	} mods;
	struct s_string *type;
} s_type;


//
// db.c
//

extern int32 next_function_id;
extern s_node *db_functions, *db_types;


//
// output.c
//


//
// x.c
//

typedef struct s_macrolist {
	char *name;
	struct s_type *(*function)(s_node *list);
} s_macrolist;

extern s_string *current_kit;


//
// var.c
//

typedef struct s_var {
	struct s_var *next;
	s_cast *var;
	uint32 offset;
	uint32 bytes;
	uint32 usage;
} s_var;

extern uint32 vars_declare_mode, vars_decl_in_progress, vars_stack_offset;
