#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


////////////////////////////////////////////////////////////////////////// GLOBAL


s_node *nodeCreate( e_node_type type ){
	s_node *list;

	list = (s_node *)malloc(sizeof(s_node));
	if(list != NULL){
		list->next = NULL;
		list->type = type;
		list->eval_type = NULL;
		list->node = NULL;
	} else {
		printf("ERROR: malloc error in nodeCreate()\n");
	}
	return(list);
}


void nodeAppendToList( s_node *list, s_node *item ){
	s_node *n;

	n = list->node;
	if(n == NULL){
		list->node = (void *)item;
	} else {
		while(n->next != NULL) n = n->next;
		n->next = item;
	}
}


s_type *nodeEvaluate( s_node *node ){
	s_num *num;
	s_var *var;

	if(node == NULL){
		printf("ERROR: no node, can't evaluate\n");
		return(NULL);
	}

	if(node->type == T_LIST_CODE){
		node->eval_type = nodeEvaluateList(node);
	} else if(node->type == T_LIST_LAMBDA){

  // TODO:

	} else if(node->type == T_LIST_DATA){

  // TODO:

	} else if(node->type == T_COMMENT){
		node->eval_type = void_type;
	} else if(node->type == T_STRING){

  // TODO:

	} else if(node->type == T_SYMBOL){

  // TODO:

	} else if(node->type == T_USTRING){
		var = varFind((s_string *)node->node);
		if(var == NULL){
			printf("ERROR: undeclared variable '%s', can't evaluate\n",
				((s_string *)node->node)->data);
			return(NULL);
		}

		if(var->bytes <= 4){
  printf("** eval 32bit variable '%s' [%s]\n", var->var->name->data, var->var->type->data);
			outputAsm("push dword [ebp-%d]\n", var->offset);
		} else if(var->bytes <= 8){
  printf("** eval 64bit variable '%s' [%s]\n", var->var->name->data, var->var->type->data);
			outputAsm("push dword [ebp-%d+4]\n", var->offset);
			outputAsm("push dword [ebp-%d]\n", var->offset);
		} else {
			printf("ERROR: can't evaluate variables greater than 8 bytes\n");
			return(NULL);
		}
		node->eval_type = typeCreate(var->var->mods.all, var->var->type->data);
		var->usage++;

	} else if(node->type == T_NUMBER){
		num = (s_num *)node->node;
		if(num->type == NUMBER_SIGNED || num->type == NUMBER_OK){
			if((num->value >> 32) > 0){
				node->eval_type = typeCreate(0, "int64");
				outputAsm("push dword 0x%X\n", (int)(num->value>>32));
				outputAsm("push dword 0x%X\n", (int)(num->value&0xFFFFFFFF));
			} else {
				node->eval_type = typeCreate(0, "int32");
				outputAsm("push dword 0x%X\n", (int)num->value);
			}
		} else {
			printf("ERROR: unsuported number, can't evaluate\n");
			return(NULL);
		}
	} else if(node->type == T_TYPE){

  // TODO:

	} else if(node->type == T_CAST){

  // TODO:

	} else {
		printf("ERROR: unknown node type, can't evaluate\n");
		return(NULL);
	}
	return(node->eval_type);
}


s_type *nodeEvaluateList( s_node *list ){
	s_node *n;
	int64 fn;

	if(list == NULL){
		printf("ERROR: no list, can't evaluate\n");
		return(NULL);
	}

	if(list->type == T_LIST_CODE || list->type == T_LIST_LAMBDA || list->type == T_LIST_DATA){
		n = list->node;
		if(n == NULL){
			printf("ERROR: empty list, cant evaluate\n");
			return(NULL);
		}
	} else n = list;

	if(n->type == T_COMMENT){
		list->eval_type = void_type;
	} else if(n->type == T_STRING){

  // TODO:

	} else if(n->type == T_SYMBOL){

  // TODO:

	} else if(n->type == T_USTRING){
		if(vars_declare_mode && !vars_decl_in_progress)
			if(!stringCompareData(((s_string *)n->node)->data, "set"))
				vars_declare_mode = 0;
		switch(callMacro((s_string *)n->node, n->next, list)){
		case 0:
			printf("ERROR: %s macro failed\n", ((s_string *)n->node)->data);
			return(NULL);
		case -1:
			printf(" function CALL %s\n", ((s_string *)n->node)->data);
			fn = callFunction((s_string *)n->node, n->next, list);
			if(fn == -1){
				printf("ERROR: %s function not found\n", ((s_string *)n->node)->data);
				return(NULL);
			} else {
				outputAsm("call f%Ld\n", fn);
			}
		}
	} else if(n->type == T_NUMBER){

  // TODO:

	} else if(n->type == T_TYPE){

  // TODO:

	} else if(n->type == T_CAST){

  // TODO:

	} else {
		printf("ERROR: unknown list type, can't evaluate\n");
		return(NULL);
	}
	return(list->eval_type);
}
