#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


void parserIndent( int32 level ){
	while(level){
		printf(" ");
		level--;
	}
}
void parserPrint( s_node *list, int32 level ){
	s_node *n;
	s_string *n_str;
	s_num *n_num;
	s_type *n_type;
	s_cast *n_cast;
	char t[100];

	if(list == NULL) return;
	n = list->node;
	while(n != NULL){
		parserIndent(level);

		if(n->eval_type != NULL){
			sprintf(t, " (%s)", n->eval_type->type->data);
		} else strcpy(t, "");

		switch(n->type){
		case T_LIST_CODE:
			printf("LIST -> CODE%s\n", t);
			parserPrint(n, level+1);
			break;
		case T_LIST_LAMBDA:
			printf("LIST -> LAMBDA%s\n", t);
			parserPrint(n, level+1);
			break;
		case T_LIST_DATA:
			printf("LIST -> DATA%s\n", t);
			parserPrint(n, level+1);
			break;
		case T_COMMENT:
			n_str = (s_string *)n->node;
			printf("COMMENT [%s]%s\n", n_str->data, t);
			break;
		case T_STRING:
			n_str = (s_string *)n->node;
			printf("STRING [%s]%s\n", n_str->data, t);
			break;
		case T_SYMBOL:
			n_str = (s_string *)n->node;
			printf("SYMBOL [%s]%s\n", n_str->data, t);
			break;
		case T_USTRING:
			n_str = (s_string *)n->node;
			printf("USTRING [%s]%s\n", n_str->data, t);
			break;
		case T_NUMBER:
			n_num = (s_num *)n->node;
			printf("NUMBER [%Ld]%s\n", n_num->value, t);
			break;
		case T_TYPE:
			n_type = (s_type *)n->node;
			printf("TYPE :[%d][%s]%s\n", n_type->mods.all,
				n_type->type->data, t);
			break;
		case T_CAST:
			n_cast = (s_cast *)n->node;
			printf("CAST [%s]:[%d][%s]%s\n", n_cast->name->data,
				n_cast->mods.all, n_cast->type->data, t);
			break;
		default:
			printf("UNKNOWN\n");
		}

		n = n->next;
	}
}


s_node *parseItem( s_string *source, uint32 *i ){
	char c;
	uint32 si;
	s_node *item;
	s_string *node_string;
	s_num *node_num;
	s_cast *node_cast;

	c = source->data[*i];
	if(c == '(' || c == '{' || c == '['){
		// sublist
		item = parseList(source, i);
		return(item);
	}

	si = *i;

	if(c == ';'){
		// comment
		if(stringFindChar(source, i, '\n')){
			node_string = stringCloneRange(source, si + 1, *i - 1);
			item = nodeCreate(T_COMMENT);
			item->node = (void *)node_string;
			return(item);
		}
	}

	if(c == '\"'){
		// quoted string
		(*i)++;
		if(stringFindChar(source, i, '\"')){
			node_string = stringCloneRange(source, si + 1, *i - 1);
			(*i)++;
			item = nodeCreate(T_STRING);
			item->node = (void *)node_string;
			return(item);
		}
	}

	c = stringFindWhite(source, i);
	node_string = stringCloneRange(source, si, *i - 1);

	if(node_string->data[0] == '\''){
		// quoted symbol
		if(node_string->size == 1){
			printf("%d: empty symbol\n", *i);
			return(NULL);
		}
		item = nodeCreate(T_SYMBOL);
		item->node = (void *)node_string;
		return(item);
	}

	node_num = stringToNumber(node_string);
	if(node_num->type <= NUMBER_OK){
		item = nodeCreate(T_NUMBER);
		item->node = (void *)node_num;
		return(item);
	}

	node_cast = castCreate(node_string);
	if(node_cast != NULL){
		if(node_cast->name->size > 0){
			item = nodeCreate(T_CAST);
			item->node = (void *)node_cast;
		} else {
			item = nodeCreate(T_TYPE);
			item->node = typeCreate(node_cast->mods.all, node_cast->type->data);
			castFree(node_cast);
		}
		return(item);
	}

	// unquoted symbol
	// possible function/variable name or string
	item = nodeCreate(T_USTRING);
	item->node = (void *)node_string;
	return(item);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


s_node *parseList( s_string *source, uint32 *i ){
	char c;
	s_node *list, *item;

	c = source->data[*i];
	if(c == '(' || c == '{' || c == '['){
		list = NULL; // satisfy gcc
		switch(c){
		case '(': list = nodeCreate(T_LIST_CODE); break;
		case '{': list = nodeCreate(T_LIST_LAMBDA); break;
		case '[': list = nodeCreate(T_LIST_DATA); break;
		}
		(*i)++;
		c = stringSkipWhite(source, i);

		while(c != ')' && c != '}' && c != ']' && c != '\0'){
			item = parseItem(source, i);
			if(item == NULL) return(NULL);
			nodeAppendToList(list, item);
			c = stringSkipWhite(source, i);
		}

		if(c == ')' && list->type == T_LIST_CODE){
		} else if(c == '}' && list->type == T_LIST_LAMBDA){
		} else if(c == ']' && list->type == T_LIST_DATA){
		} else {
			printf("%d: incorrect closing bracket\n", *i);
			return(NULL);
		}
		(*i)++;

		return(list);
	} else {
		printf("%d: found garbage\n", *i);
		return(NULL);
	}
}


s_node *parseSource( s_string *source ){
	uint32 i;
	s_node *mainlist, *list;

	i = 0;
	mainlist = nodeCreate(T_LIST_OUTTER);
	stringSkipWhite(source, &i);
	while(i != source->size){
		list = parseList(source, &i);
		if(list == NULL) break;
		nodeAppendToList(mainlist, list);
		if(stringSkipWhite(source, &i) == '\0') break;
	}

  printf("----------------------- parsed nodes\n");
  parserPrint(mainlist, 0);

	return(mainlist);
}
