#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global

// local

////////////////////////////////////////////////////////////////////////// PRIVATE


////////////////////////////////////////////////////////////////////////// GLOBAL


s_string *stringCreate( uint32 size ){
	s_string *s;

	s = (s_string *)malloc(sizeof(s_string));
	if(s != NULL){
                s->data = (char *)malloc(size+1);
		if(s->data == NULL){
			printf("ERROR: malloc error in stringCreate()\n");
			free(s);
			return(NULL);
		}
                s->size = size;
	} else {
		printf("ERROR: malloc error in stringCreate()\n");
	}
	return(s);
}


void stringFree( s_string *s ){
	if(s != NULL){
		if(s->data != NULL) free(s->data);
		free(s);
	}
}


s_string *stringToLower( s_string *s ){
	char *cs, *cd;
	s_string *d;

	d = stringCreate(s->size);
	cd = d->data;
	cs = s->data;
	while(*cs != '\0'){
		if(*cs >= 'A' && *cs <= 'Z') *cd = *cs | 32;
		else *cd = *cs;
		cs++;
		cd++;
	}
	*cd = '\0';
	return(d);
}


char stringSkipWhite( s_string *s, uint32 *index ){
	// skips over any whitespace: ' ', '\t', '\n'.
	// if current character is not whitespace then this does nothing.
	// index points to first non-whitespace character
	char c;
	uint32 i;

	i = *index;
	do{
		c = s->data[i];
		if(c == '\0') return('\0');
		i++;
	} while( (c==' ') || (c=='\t') || (c=='\n') );
	*index = i - 1;
	return(c);
}


char stringFindWhite( s_string *s, uint32 *index ){
	// skip over non-whitespace, ')' counts as whitespace.
	// if current character is whitespace then this does nothing.
	// index points to last non-whitespace character
	char c;
	uint32 i;

	i = *index;
	do{
		c = s->data[i];
		if(c == '\0') return('\0');
		i++;
	} while( (c!=' ') && (c!='\t') && (c!='\n') && (c!=')') && (c!='}') && (c!=']') );
	*index = i - 1;
	return(c);
}


char stringExpect( s_string *s, uint32 *index, char e ){
	// skips whitespace and compare character.
	// if match, index points to character after expected character, return expected character
	// if not, index points to unexpected character, return NULL
	char c;

	c = stringSkipWhite(s, index);
	if(c == e){
		(*index)++;
		return(e);
	}
	return('\0');
}


char stringFindChar( s_string *s, uint32 *index, char m ){
	// advance until character found
	// index points to character found
	char c;
	uint32 i;

	i = *index;
	do{
		c = s->data[i];
		if(c == '\0') return('\0');
		i++;
	} while(c != m);
	*index = i - 1;
	return(c);
}


s_string *stringCloneRange( s_string *s, uint32 start, uint32 end ){
	uint32 z;
	s_string *s2;

	if(start >= s->size || end >= s->size || start > end) return(NULL);

	s2 = stringCreate(end-start+1);
	if(s2 != NULL){
		for(z = 0; start <= end; start++){
			s2->data[z] = s->data[start];
			z++;
		}
		s2->data[z] = '\0';
	}
	return(s2);
}


int32 stringCompare( s_string *s1, s_string *s2 ){
	// case-insensitive string comparison
	// return 1 if match, else 0
	char c1, c2, *sd1, *sd2;

	sd1 = s1->data;
	sd2 = s2->data;
	while(*sd1 != '\0' && *sd2 != '\0'){
		c1 = *sd1;
		c2 = *sd2;
		if(c1 >= 'A' && c1 <= 'Z') c1 |= 32;
		if(c2 >= 'A' && c2 <= 'Z') c2 |= 32;
		if(c1 != c2) return(0);
		sd1++;
		sd2++;
	}
	if(*sd1 != *sd2) return(0);
	return(1);
}


int32 stringCompareData( char *s1, char *s2 ){
	// case-insensitive string comparison
	// return 1 if match, else 0
	char c1, c2;

	while(*s1 != '\0' && *s2 != '\0'){
		c1 = *s1;
		c2 = *s2;
		if(c1 >= 'A' && c1 <= 'Z') c1 |= 32;
		if(c2 >= 'A' && c2 <= 'Z') c2 |= 32;
		if(c1 != c2) return(0);
		s1++;
		s2++;
	}               
	if(*s1 != *s2) return(0);
	return(1);
}
