#include "common.h"
#include "functions.h"


//////////////////////////////////////////////////////////////////////////

// global
uint32 vars_declare_mode, vars_decl_in_progress, vars_stack_offset;

// local
s_var *variables;

////////////////////////////////////////////////////////////////////////// PRIVATE


////////////////////////////////////////////////////////////////////////// GLOBAL


s_var *varCreate( s_cast *var ){
	s_var *v;

	v = (s_var *)malloc(sizeof(s_var));
	if(v != NULL){
		v->var = var;
		v->next = variables;
		v->usage = 0;
		variables = v;

		// compute size of type
		if(stringCompareData(var->type->data, "int8")){
			v->bytes = 1;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "int16")){
			v->bytes = 2;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "int32")){
			v->bytes = 4;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "int")){
			v->bytes = 4;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "int64")){
			v->bytes = 8;
			vars_stack_offset += 8;
		} else if(stringCompareData(var->type->data, "uint8")){
			v->bytes = 1;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "uint16")){
			v->bytes = 2;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "uint32")){
			v->bytes = 4;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "uint")){
			v->bytes = 4;
			vars_stack_offset += 4;
		} else if(stringCompareData(var->type->data, "uint64")){
			v->bytes = 8;
			vars_stack_offset += 8;
		}else {
			printf("ERROR: unknown variable type\n");
			return(NULL);
		}

		// compute location on stack
		v->offset = vars_stack_offset;
	} else {
		printf("ERROR: malloc error in varCreate()\n");
	}
	return(v);
}


void varFreeAll( ){
	s_var *v, *vo;

	v = variables;
	while(v != NULL){
		vo = v->next;
		castFree(v->var);
		free(v);
		v = vo;
	}
	variables = NULL;
	vars_stack_offset = 0;
}


s_var *varFind( s_string *name ){
	s_var *v;

	v = variables;
	while(v != NULL){
		if(stringCompare(name, v->var->name)){
			return(v);
		}
		v = v->next;
	}
	return(NULL);
}


void varUsageWarnings( ){
	s_var *v;

	v = variables;
	while(v != NULL){
		if(v->usage == 0)
			printf("WARNING: variable '%s' was not used\n", v->var->name->data);
		v = v->next;
	}
}
