#include "common.h"
#include "functions.h"

// prototypes for private functions defined below, for macro list
s_type *funcSet( s_node *list );
s_type *funcDefun( s_node *list );
s_type *funcDefunRaw( s_node *list );
s_type *funcDefType( s_node *list );
s_type *funcSlot( s_node *list );
s_type *funcDefault( s_node *list );
s_type *funcCastToParent( s_node *list );
s_type *funcDefMacro( s_node *list );
s_type *funcReturn( s_node *list );
s_type *funcCastReturn( s_node *list );
s_type *funcExec( s_node *list );
s_type *funcAsm( s_node *list );
s_type *funcPlatform( s_node *list );
s_type *funcIf( s_node *list );
s_type *funcToInt64( s_node *list );
s_type *funcAddIntInt( s_node *list );


//////////////////////////////////////////////////////////////////////////

// global

// local
s_macrolist macros[] = {
	{"set", &funcSet},
	{"defun", &funcDefun},
	{"defun-raw", &funcDefunRaw},
	{"deftype", &funcDefType},
	{"slot", &funcSlot},
	{"default", &funcDefault},
	{"cast-to-parent", &funcCastToParent},
	{"defmacro", &funcDefMacro},
	{"return", &funcReturn},
	{"cast-return", &funcCastReturn},
	{"exec", &funcExec},
	{"asm", &funcAsm},
	{"platform", &funcPlatform},
	{"if", &funcIf},
	{"to-int64", &funcToInt64},
	{"add", &funcAddIntInt},
	{NULL, NULL}
};
char base_ebp[] = "ebp-";
char base_none[] = "";
uint32 defun, defmacro, deftype, label_id;

////////////////////////////////////////////////////////////////////////// PRIVATE


s_type *funcSet( s_node *list ){
	char *base;
	s_cast *node_cast;
	s_var *v;
	s_type *input;
	s_node *temp;
	uint32 count;

	printf(" calling SET\n");

	switch(list->type){
	case T_CAST:
		if(!vars_declare_mode){
			printf("ERROR: declarations only allowed at top of code\n");
			return(NULL);
		}
		node_cast = list->node;
		if(varFind(node_cast->name) != NULL){
			printf("ERROR: variable already declared\n");
			return(NULL);
		}
		v = varCreate(node_cast);
		if(v == NULL) return(NULL);
		vars_decl_in_progress = 1;
		break;
	case T_USTRING:
		v = varFind((s_string *)list->node);
		if(v == NULL){
			printf("ERROR: undeclared variable '%s'\n",
				((s_string *)list->node)->data);
			return(NULL);
		}
		break;
	default:
		printf("ERROR: invalid first parameter to SET\n");
		return(NULL);
	}

	input = void_type;
	count = 0;
	if(list->next != NULL){
		temp = list->next;
		while(temp != NULL){
			input = nodeEvaluate(temp);
			if(input == NULL) return(NULL);
			if(input != void_type) count++;
			temp = temp->next;
		}

		if(count > 1){
			printf("ERROR: multiple values passed to SET are unsuported at this time\n");
			return(NULL);
		}

		temp = list->next;
		do{
			input = temp->eval_type;
			temp = temp->next;
		} while(temp != NULL && input == void_type);
	}

	if(count > 0){

  // use type comparison function for this
		if(!stringCompare(v->var->type, input->type)){
			printf("ERROR: value type [%s] does not match type [%s] of variable '%s'\n",
				input->type->data, v->var->type->data, v->var->name->data);
			return(NULL);
		}

		if(v->var->mods.x.pointer) base = base_none;
		else base = base_ebp;

		outputAsm("pop eax\n");
		switch(v->bytes){ // destination is global object
		case 1:
			outputAsm("mov byte [%s%d], al\n", base, v->offset);
			break;
		case 2:
			outputAsm("mov word [%s%d], ax\n", base, v->offset);
			break;
		case 4:
			outputAsm("mov dword [%s%d], eax\n", base, v->offset);
			break;
		case 8:
			outputAsm("pop ebx\n");
			outputAsm("mov dword [%s%d], eax\n", base, v->offset);
			outputAsm("mov dword [%s%d+4], ebx\n", base, v->offset);
			break;
		}
	} else {

  // flag variable as uninitialized
  printf("WARNING: variable '%s' has been declared but not initialized\n", v->var->name->data);

	}

	vars_decl_in_progress = 0;
	return(void_type);
}


s_type *funcDefun( s_node *list ){
	s_node *inputs, *body;
	s_string *kit, *desc;
	s_node *func, *node, *node2, *interface, *regfunc;
	uint32 private;
	char out[1000];

	// ( DEFUN private kit:string name:cast inputs:castlist desc:string body:[codelist )

	// private -- optional keyword used to create private functions
	// kit -- namespace for function name
	// name -- function name and return type, valid characters for name are 
	//	[a-zA-Z0-9~!@#$%^&*-_+=|\></?]*
	// inputs -- list of casts for each function input
	// desc -- function description
	// body -- multiple code lists defining function body

	if(defun) return(NULL);
	printf("called DEFUN\n");

	// get private/kit
	if(list == NULL){ printf(" empty DEFUN\n"); return(NULL); }
	if(list->type == T_USTRING){
		kit = (s_string *)list->node;
		if(stringCompareData(kit->data, "private")){
			stringFree(kit);
			private = 1;
			list = list->next;
			if(list == NULL){ printf(" no module name\n"); return(NULL); }
			if(list->type == T_USTRING)
				kit = (s_string *)list->node;
		} else private = 0;
	} else { printf(" no module name\n"); return(NULL); }

	// get name and return type
	list = list->next;
	if(list == NULL){ printf(" empty DEFUN after module name\n"); return(NULL); }
	if(list->type == T_CAST){
		func = list;
	} else if(list->type == T_LIST_CODE && ((s_node *)list->node)->type == T_CAST){
		func = list;
	} else { printf(" no function name\n"); return(NULL); }

	// get inputs
	list = list->next;
	if(list == NULL){ printf(" empty DEFUN after function name\n"); return(NULL); }
	if(list->type == T_LIST_CODE){
		inputs = (s_node *)list->node;
	} else { printf(" no parameter list\n"); return(NULL); }

	// get description
	list = list->next;
	if(list == NULL){ printf(" empty DEFUN after parameter list\n"); return(NULL); }
	if(list->type == T_STRING){
		desc = (s_string *)list->node;
	} else { printf(" no description\n"); return(NULL); }

	// get body
	list = list->next;
	if(list == NULL){ printf(" empty DEFUN after description\n"); return(NULL); }
	if(list->type == T_LIST_CODE){
		body = list;
	} else { printf(" no function body\n"); return(NULL); }


	// check for interface collisions with other functions
	node = NULL;
	while(1){
		node = dbFindFunction(((s_cast *)func->node)->name, node);
		if(node == NULL) break;

		interface = dbGetInterface(node);
		if(interface == NULL){
			printf("ERROR: unable to get function interface\n");
			return(NULL);
		}

		if(dbCompareRawInterfaces(inputs, (s_node *)interface->node)){
			printf("ERROR: function '%s' exists in dictionary with same interface\n",
				((s_cast *)func->node)->name->data);
			return(NULL);
		}
	}


	// build function interface string
regfunc = nodeCreate(T_LIST_CODE);
node = nodeCreate(T_NUMBER);
node->node = (void *)malloc(sizeof(s_num));
((s_num *)node->node)->value = (int64)next_function_id;
nodeAppendToList(regfunc, node);
	sprintf(out, ";F(%d ", next_function_id);		// tag and function ID
	if(private){
node = nodeCreate(T_USTRING);
node->node = (void *)stringCreate(7);
strcpy(((s_string *)node->node)->data, "private");
nodeAppendToList(regfunc, node);
		strcat(out, "private ");			// private
	}
node = nodeCreate(T_USTRING);
node->node = (void *)kit;
nodeAppendToList(regfunc, node);
	strcat(out, kit->data);					// module
	strcat(out, " ");
node = nodeCreate(T_CAST);
node->node = func->node;
nodeAppendToList(regfunc, node);
	castDisplay(out, func);					// function name and rtype
	strcat(out, "(");
node = nodeCreate(T_LIST_CODE);
nodeAppendToList(regfunc, node);
	while(inputs != NULL){
		if(castDisplay(out, inputs) == 0) return(NULL);	// parameters
node2 = nodeCreate(T_CAST);
node2->node = inputs->node;
nodeAppendToList(node, node2);
		inputs = inputs->next;
	}
	strcat(out, ") \"");
node = nodeCreate(T_STRING);
node->node = (void *)desc;
nodeAppendToList(regfunc, node);
	strcat(out, desc->data);				// description
	strcat(out, "\")\n");
	outputInterface(out);
nodeAppendToList(db_functions, regfunc);


	// evaluate body
	defun = 1;
	label_id = 0;
	vars_declare_mode = 1;
	vars_decl_in_progress = 0;
	vars_stack_offset = 0;
	varFreeAll();
	while(body != NULL){
		if(nodeEvaluateList(body) == NULL){
			defun = 0;
			return(NULL);
		}
		body = body->next;
	}

	// success
	varUsageWarnings();
	next_function_id++;
	defun = 0;
	outputFlush();
	return(void_type);
}


s_type *funcDefunRaw( s_node *list ){
	// ( DEFUN-RAW name:cast inputs:castlist desc:string body:[codelist )
	// ( DEFUN-RAW kit:string name:cast inputs:castlist desc:string body:[codelist )

	// kit -- namespace for function name
	// name -- function name and return type, valid characters for name are 
	//	[a-zA-Z0-9~!@#$%^&*-_+=|\></?]*
	// inputs -- list of casts for each function input
	// desc -- function description
	// body -- multiple code lists defining function body

	if(defun) return(NULL);

	printf("called DEFUN-RAW\n");
	defun = 1;

  // body

	defun = 0;
	return(NULL);
}


s_type *funcDefType( s_node *list ){
	// ( DEFTYPE private name:cast parameters:castlist desc:string body:[codelist )
	// ( DEFTYPE private kit:string name:cast parameters:castlist desc:string body:[codelist )

	// private -- optional keyword used to create private types
	// kit -- namespace for function name
	// name -- type name and parent type, valid characters for name are
	//	[a-zA-Z][a-zA-Z0-9_-]*
	// parameters -- list of casts for each function input
	// desc -- function description
	// body -- multiple code lists defining function body 

	if(defun) return(NULL);

	printf("called DEFTYPE\n");
	defun = 1; deftype = 1;

  // body

	defun = 0; deftype = 0;
	return(NULL);
}


s_type *funcSlot( s_node *list ){
	// ; name pattern [a-zA-Z][a-zA-Z0-9_-]*
	// ( SLOT readonly private name:cast default-value:? )

	if(!deftype) return(NULL);

	printf("called SLOT\n");
	return(NULL);
}


s_type *funcDefault( s_node *list ){
	// ; name pattern [a-zA-Z][a-zA-Z0-9_-]*
	// ( DEFAULT private name:cast )

	if(!deftype) return(NULL);

	printf(" called DEFAULT\n");
	return(NULL);
}


s_type *funcCastToParent( s_node *list ){
	// ( CAST-TO-PARENT object:? )

	if(!defun) return(NULL);

	printf(" called CAST-TO-PARENT\n");
	return(NULL);
}


s_type *funcDefMacro( s_node *list ){
	// ( DEFMACRO private name:cast inputs:castlist desc:string body:[codelist )
	// ( DEFMACRO private kit:string name:cast inputs:castlist desc:string body:[codelist )

	// private -- optional keyword used to create private macros
	// kit -- namespace for macro name
	// name -- macro name and return type, valid characters for name are 
	//	[a-zA-Z0-9~!@#$%^&*-_+=|\></?]
	// inputs -- list of casts for each macro input
	// desc -- macro description
	// body -- multiple code lists defining macro body 

	if(defun) return(NULL);

	printf("called DEFMACRO\n");
	defun = 1; defmacro = 1;

  // body

	defun = 0; defmacro = 0;
	return(NULL);
}


s_type *funcReturn( s_node *list ){
	// ( RETURN value:? )
	// ( RETURN )

	if(!defun) return(NULL);

	if(vars_declare_mode){
		printf("ERROR: can't return while in variable declaration mode\n");
		return(NULL);
	}

	printf(" called RETURN\n");

	if(vars_stack_offset)
		outputAsm("add esp, %d\n", vars_stack_offset);
	outputAsm("pop ebp\n");
	outputAsm("ret\n");
	return(void_type);
}


s_type *funcCastReturn( s_node *list ){
	// ( CAST-RETURN value:? )

	if(!defun) return(NULL);

	printf(" called CAST-RETURN\n");
	return(NULL);
}

s_type *funcExec( s_node *list ){
	// ( EXEC body:[codelist )

	if(!defun) return(NULL);

	printf(" called EXEC\n");
	return(NULL);
}


s_type *funcAsm( s_node *list ){
	// ( ASM platform:ustring body:[codelist )

	if(!defun) return(NULL);

	printf(" called ASM\n");
	return(NULL);
}


s_type *funcPlatform( s_node *list ){
	// ( PLATFORM platform:ustring body:[codelist )
	// ( PLATFORM platform:#ustring body:[codelist )

	if(!defun) return(NULL);

	printf(" called PLATFORM\n");
	return(NULL);
}


s_type *funcIf( s_node *list ){
	s_node *false;
	s_type *value;
	uint32 lid;

	// ( IF condition:bool true:codelist)
	// ( IF condition:bool true:codelist false:codelist)

	if(!defun) return(NULL);

	printf(" called IF\n");

	lid = label_id;
	label_id++;

	if(list == NULL){
		printf("ERROR: condition is missing\n");
		return(NULL);
	}
	if(list->next == NULL){
		printf("ERROR: true block is missing\n");
		return(NULL);
	}

	value = nodeEvaluate(list);			// condition
	if(!stringCompareData(value->type->data, "int32")){
		printf("ERROR: condition is not an int value\n");
		return(NULL);
	}
	false = list->next->next;

	outputAsm("pop eax\n");
	outputAsm("xor eax,eax\n");
	outputAsm("jnz .then_%d\n", lid);

	if(false != NULL) outputAsm("jmp .else_%d\n", lid);
	else outputAsm("jmp .endif_%d\n", lid);

	outputAsm(".then_%d:\n", lid);
	if(nodeEvaluateList(list->next) == NULL)	// true block
		return(NULL);

	if(false != NULL){
		outputAsm("jmp .endif_%d\n", lid);
		outputAsm(".else_%d:\n", lid);

		if(nodeEvaluateList(false) == NULL)	// false block
			return(NULL);
	}
	outputAsm(".endif_%d:\n", lid);

	return(void_type);
}


s_type *funcToInt64( s_node *list ){
	s_type *value;

	// (TO-INT64 n:int32 )
	// (TO-INT64 n:int16 )
	// (TO-INT64 n:int8 )

	if(!defun) return(NULL);

	printf(" calling TO-INT64\n");

	value = nodeEvaluate(list);
	if(value == NULL) return(NULL);

	if(!stringCompareData(value->type->data, "int32"))
	  if(!stringCompareData(value->type->data, "int16"))
	    if(!stringCompareData(value->type->data, "int8")){
		printf("ERROR: can't convert %s to int64\n", value->type->data);
		return(NULL);
	    }

	outputAsm("pop eax\n");		// fix stack
	outputAsm("push dword 0\n");
	outputAsm("push eax\n");
	return(typeCreate(0, "int64"));
}


s_type *funcAddIntInt( s_node *list ){
	s_type *value;

	// (ADD a:int32 b:int32)

	if(!defun) return(NULL);

	printf(" calling ADD\n");

	value = nodeEvaluate(list);
	if(value != NULL){
	  if(stringCompareData(value->type->data, "int32")){
	    value = nodeEvaluate(list->next);
	    if(value != NULL){
	      if(stringCompareData(value->type->data, "int32")){
		if(list->next->next == NULL){
			outputAsm("pop ebx\n");
			outputAsm("pop eax\n");
			outputAsm("add eax, ebx\n");
			outputAsm("push eax\n");
			return(typeCreate(0, "int32"));
		}
	      }
	    }
	  }
	}

//	callFunction();
	return(NULL);
}


////////////////////////////////////////////////////////////////////////// GLOBAL


int32 callMacro( s_string *name, s_node *p1, s_node *list ){
	s_string *s;
	int32 m;

	s = stringToLower(name);

	m = 0;
	while(macros[m].name != NULL){
		if(stringCompareData(s->data, macros[m].name)){
			stringFree(s);
			list->eval_type = (*macros[m].function)(p1);
			if(list->eval_type == NULL) return(0);
			else return(1);
		}
		m++;
	}

	stringFree(s);
	return(-1);
}


int64 callFunction( s_string *name, s_node *p1, s_node *list ){
	s_string *s;
	s_node *func, *interface, *p;

	// eval all nodes in list
	p = p1;
	while(p != NULL){
		if(nodeEvaluate(p) == NULL) return(-1);
		p = p->next;
	}

	s = stringToLower(name);

  // check name for kit.func

	func = NULL;
	while(1){
		func = dbFindFunction(name, func);
		if(func == NULL) break;

		interface = dbGetInterface(func);
		if(interface == NULL) break;

		if(dbCompareInterfaces(p1, (s_node *)interface->node)){
			list->eval_type = dbGetReturnType(func);
			return(((s_num *)((s_node *)func->node)->node)->value);
		}
	}
	return(-1);
}
